# Code based on https://github.com/alts/karel
#-*- coding: utf-8 -*-
from __future__ import print_function

import re
import numpy as np
from collections import Counter

from .hero import Hero
from .utils import Tcolors, get_rng

def draw2d(array):
    print("\n".join(["".join(["#" if val > 0 else "." for val in row]) for row in array]))

def border_mask(array, value):
    array[0,:], array[-1,:], array[:,0], array[:,-1] = value, value, value, value

def hero_action(func):
    def fn(*args, **kwargs):
        self = args[0]
        out = func(self)
        if self.debug:
            print(func.__doc__, out)
            self.draw()
        return out
    return fn

def marker_action(func):
    def fn(*args, **kwargs):
        self = args[0]
        out = func(self)
        if self.debug:
            print(func.__doc__, out)
        return out
    return fn

world_condition = marker_action


class Karel(object):
    HERO_CHARS = '<^>v'
    MARKER_CHAR = 'o'
    WALL_CHAR = '#'
    EMPTY_CHAR = '.'

    COLOR_DICT = {
            WALL_CHAR: WALL_CHAR,
            MARKER_CHAR: Tcolors.WARNING + MARKER_CHAR + Tcolors.ENDC,
    }
    #COLOR_DICT.update({
    #        str(num): Tcolors.WARNING + str(num) + Tcolors.ENDC for num in range(9) })
    COLOR_DICT.update({
            char: Tcolors.OKGREEN + char + Tcolors.ENDC for char in HERO_CHARS })

    def __init__(
            self, world_size=None, world_path=None, rng=None,
            max_marker_in_cell=1, debug=False):

        self.debug = debug
        self.rng = get_rng(rng)

        self.markers = []
        if world_path is not None:
            self.parse_world(world_path)
        elif world_size is not None:
            self.random_world(world_size, max_marker_in_cell)
        else:
            raise Exception(" [!] one of `world_size` and `world_path` should be passed")

        state = np.zeros_like(self.world, dtype=np.int8)
        self.zero_state = np.tile(np.expand_dims(state, -1), [1, 1, 16])

        if self.debug: self.draw()

    def __enter__(self):
        self.start_screen()
        return self

    def __exit__(self, *args):
        self.end_screen()

    def start_screen(self):
        pass

    def end_screen(self):
        pass

    def random_world(self, world_size, max_marker_in_cell, wall_ratio=0.2, marker_ratio=0.3):
        height, width = world_size

        if height <= 2 or width <= 2:
            raise Exception(" [!] `height` and `width` should be larger than 2")

        # blank world
        self.world = np.chararray((height, width))
        self.world[:] = "."

        # wall
        wall_array = self.rng.rand(height, width)

        self.world[wall_array < wall_ratio] = self.WALL_CHAR
        border_mask(self.world, self.WALL_CHAR)

        # hero
        x, y, direction = self.rng.randint(1, width-1), \
                self.rng.randint(1, height-1), self.rng.randint(4)
        self.hero = Hero((x, y), ((-1, 0), (1, 0), (0, -1), (0, 1))[direction])
        self.world[y, x] = '.'

        # markers
        marker_array = self.rng.rand(height, width)
        marker_array = (wall_array >= wall_ratio) & (marker_array < marker_ratio)
        border_mask(marker_array, False)

        self.markers = []
        for (y, x) in zip(*np.where(marker_array > 0)):
            self.markers.append((x, y))

        self.world = self.world.astype(str).tolist()

    def parse_world(self, world_path):
        directions = {
            '>': (1, 0),
            '^': (0, 1),
            '<': (-1, 0),
            'v': (0, -1),
        }

        world = [[]]
        with open(world_path) as f:
            for y, line in enumerate(f):
                row = []
                for x, char in enumerate(line.strip()):
                    if char in self.HERO_CHARS:
                        self.hero = Hero((x + 1, y + 1), directions[char])
                        char = '.'
                    elif char == self.MARKER_CHAR:
                        self.markers.append((x + 1, y + 1))
                        char = '.'
                    elif char.isdigit():
                        for _ in range(int(char)):
                            self.markers.append((x + 1, y + 1))
                        char = '.'
                    elif char in [self.WALL_CHAR, self.EMPTY_CHAR]:
                        pass
                    else:
                        raise Exception(" [!] `{}` is not a valid character".format(char))
                    row.append(char)
                world.append([self.WALL_CHAR] + row + [self.WALL_CHAR])

        world.append([])
        for _ in range(len(world[1])):
            world[0].append(self.WALL_CHAR)
            world[-1].append(self.WALL_CHAR)

        self.world = world

    def draw(self, prefix=None, skip_number=True, with_color=True):
        canvas = np.array(self.world)

        for (x, y), count in Counter(self.markers).items():
            canvas[y][x] = str(count)

        canvas[self.hero.position[1]][self.hero.position[0]] = self.hero_char()

        for idx, row in enumerate(canvas):
            row_text = "".join(row)
            if skip_number:
                row_text = re.sub('\d', self.MARKER_CHAR, row_text)

            if idx == 0:
                if with_color:
                    text = "{}{}{}{}".format(
                            Tcolors.OKBLUE, prefix, Tcolors.ENDC, row_text)
                else:
                    text = "{}{}".format(prefix, row_text)
            else:
                text = "{}{}".format(len(prefix) * " ", row_text)

            if with_color:
                text = re.sub('.'.format(self.WALL_CHAR), lambda x: self._color_fn(x), text)

            print(text)

    def _color_fn(self, x):
        char = x.group()
        if char in self.COLOR_DICT:
            return self.COLOR_DICT[char]
        else:
            return char

    @property
    def state(self):
        """
            0: Hero facing North
            1: Hero facing South
            2: Hero facing West
            3: Hero facing East
            4: Wall
            5: 0 marker
            6: 1 marker
            7: 2 marker
            8: 3 marker
            9: 4 marker
            10: 5 marker
            11: 6 marker
            12: 7 marker
            13: 8 marker
            14: 9 marker
            15: 10 marker
        """
        state = self.zero_state.copy()
        state[:,:,5] = 1

        # 0 ~ 3: Hero facing North, South, West, East
        x, y = self.hero.position
        state[y, x, self.facing_idx] = 1

        # 4: wall or not
        for jdx, row in enumerate(self.world):
            for idx, char in enumerate(row):
                if char == self.WALL_CHAR:
                    state[jdx][idx][4] = 1
                elif char == self.WALL_CHAR or char in self.HERO_CHARS:
                    state[:,:,5] = 1

        # 5 ~ 15: marker counter
        for (x, y), count in Counter(self.markers).items():
            state[y][x][5] = 0
            state[y][x][5 + count] = 1

        # draw2d(state[:,:,5])
        return state

    def draw_exception(self, exception):
        pass

    def hero_char(self):
        # index will be in (-2, -1, 1, 2)
        index = self.hero.facing[0] + 2*self.hero.facing[1]
        return ' >v^<'[index]

    @hero_action
    def move(self):
        '''Move'''
        success = True
        if not self._front_is_clear():
            #raise Exception('can\'t move. There is a wall in front of Hero')
            success = False
        else:
            self.hero.move()
        return success

    @hero_action
    def turn_left(self):
        '''Turn left'''
        self.hero.turn_left()

    @marker_action
    def turn_right(self):
        '''Turn right'''
        self.hero.turn_right()

    @marker_action
    def pick_marker(self):
        '''Pick marker'''
        position = self.hero.position
        for i, coord in enumerate(self.markers):
            if coord == self.hero.position:
                del self.markers[i]
                self.hero.pick_marker()
                break
        else:
            #raise Exception('can\'t pick marker from empty location')
            pass

    @marker_action
    def put_marker(self):
        '''Put marker'''
        if not self.hero.holding_markers():
            #raise Exception('can\'t put marker. Hero has none')
            pass
        else:
            self.markers.append(self.hero.position)
            self.hero.put_marker()
        return Counter(self.markers)[self.hero.position]

    @world_condition
    def front_is_clear(self):
        '''Check front is clear'''
        return self._front_is_clear()

    def _front_is_clear(self):
        next_x = self.hero.position[0] + self.hero.facing[0]
        next_y = self.hero.position[1] + self.hero.facing[1]
        return self.world[next_y][next_x] == '.'

    @world_condition
    def left_is_clear(self):
        '''Check left is clear'''
        return self._left_is_clear()

    def _left_is_clear(self):
        next_x = self.hero.position[0] + self.hero.facing[1]
        next_y = self.hero.position[1] - self.hero.facing[0]
        return self.world[next_y][next_x] == '.'

    @world_condition
    def right_is_clear(self):
        '''Check right is clear'''
        return self._right_is_clear()

    def _right_is_clear(self):
        next_x = self.hero.position[0] - self.hero.facing[1]
        next_y = self.hero.position[1] + self.hero.facing[0]
        return self.world[next_y][next_x] == '.'

    @world_condition
    def markers_present(self):
        '''Check markers present'''
        return self.hero.position in self.markers

    @world_condition
    def no_markers_present(self):
        '''Check no markers present'''
        return self.hero.position not in self.markers

    @property
    def facing_north(self):
        return self.hero.facing[1] == -1

    @property
    def facing_south(self):
        return self.hero.facing[1] == 1

    @property
    def facing_east(self):
        return self.hero.facing[0] == 1

    @property
    def facing_west(self):
        return self.hero.facing[0] == -1

    @property
    def facing_idx(self):
        if self.facing_north:
            return 0
        elif self.facing_south:
            return 1
        elif self.facing_east:
            return 2
        elif self.facing_west:
            return 3



# Code based on https://github.com/alts/karel

class Hero(object):
    def __init__(self, position, facing, marker_bag=None):
        self.position = position
        self.facing = facing
        self.marker_bag = None

    def move(self):
        self.position = (
            self.position[0] + self.facing[0],
            self.position[1] + self.facing[1]
        )

    def turn_left(self):
        self.facing = (
            self.facing[1],
            -self.facing[0]
        )

    def turn_right(self):
        self.turn_left()
        self.turn_left()
        self.turn_left()

    def holding_markers(self):
        return (self.marker_bag is None) or self.marker_bag > 0

    def pick_marker(self):
        if self.marker_bag is not None:
            self.marker_bag += 1

    def put_marker(self):
        if self.marker_bag is not None:
            self.marker_bag -= 1



from .parser import KarelParser
from .utils import str2bool, makedirs, pprint, beautify, TimeoutError

from datetime import datetime

__copyright__ = 'Copyright 2015 - {} by Taehoon Kim'.format(datetime.now().year)
__version__ = '1.1.2'
__license__ = 'BSD'
__author__ = 'Taehoon Kim'
__email__ = 'carpedm20@gmail.com'
__source__ = 'https://github.com/carpedm20/karel'
__description__ = 'Karel dataset for program synthesis and program induction'



from __future__ import print_function

import ply.lex as lex

from . import yacc
from .karel import Karel
from .utils import pprint, timeout, get_rng

def dummy():
    pass

class Parser(object):
    """
    Base class for a lexer/parser that has the rules defined as methods
    """
    tokens = ()
    precedence = ()

    def __init__(self, **kwargs):
        self.names = {}
        self.debug = kwargs.get('debug', 0)

        # Build the lexer and parser
        self.lexer = lex.lex(module=self, debug=self.debug)
        _, self.grammar = yacc.yacc(
                module=self, debug=self.debug,
                tabmodule='_parsetab', with_grammar=True)

        self.prodnames = self.grammar.Prodnames

    def lex_to_idx(self, code):
        tokens = []
        self.lexer.input(code)
        while True:
            tok = self.lexer.token()
            if not tok: 
                break
            idx = self.token_to_idx[tok.type]
            tokens.append(idx)
        return tokens

class KarelParser(Parser):

    tokens = [
            'DEF', 'RUN', 
            'LPAREN', 'RPAREN', 'LBRACE', 'RBRACE', 'SEMI', 'INT', #'NEWLINE',
            'WHILE', 'REPEAT',
            'IF', 'IFELSE', 'ELSE',
            'FRONT_IS_CLEAR', 'LEFT_IS_CLEAR', 'RIGHT_IS_CLEAR',
            'MARKERS_PRESENT', 'NO_MARKERS_PRESENT', 'NOT',
            'MOVE', 'TURN_RIGHT', 'TURN_LEFT',
            'PICK_MARKER', 'PUT_MARKER',
    ]
    idx_to_token = { idx: token for idx, token in enumerate(tokens) }
    token_to_idx = { token:idx for idx, token in idx_to_token.items() }

    t_ignore =' \t\n'
    t_LPAREN = r'\('
    t_RPAREN = r'\)'
    t_LBRACE = r'\{'
    t_RBRACE = r'\}'
    t_SEMI   = r';'

    t_DEF = 'def'
    t_RUN = 'run'
    t_WHILE = 'while'
    t_REPEAT = 'repeat'
    t_IF = 'if'
    t_IFELSE = 'ifelse'
    t_ELSE = 'else'
    t_NOT = 'not'

    t_FRONT_IS_CLEAR = 'front_is_clear'
    t_LEFT_IS_CLEAR = 'left_is_clear'
    t_RIGHT_IS_CLEAR = 'right_is_clear'
    t_MARKERS_PRESENT = 'markers_present'
    t_NO_MARKERS_PRESENT = 'no_markers_present'

    t_MOVE = 'move'
    t_TURN_RIGHT = 'turn_right'
    t_TURN_LEFT = 'turn_left'
    t_PICK_MARKER = 'pick_marker'
    t_PUT_MARKER = 'put_marker'


    def __init__(self, rng=None, min_int=0, max_int=19, debug=False, **kwargs):
        super(KarelParser, self).__init__(**kwargs)

        self.debug = debug
        self.min_int = min_int
        self.max_int = max_int

        self.rng = get_rng(rng)

    #########
    # lexer
    #########

    def t_INT(self, t):
        r'\d+'

        value = int(t.value)
        if not (self.min_int <= value <= self.max_int):
            raise Exception(" [!] Out of range ({} ~ {}): `{}`". \
                    format(self.min_int, self.max_int, value))

        t.value = value
        return t

    def random_INT(self):
        return self.rng.randint(self.min_int, self.max_int + 1)

    def t_error(self, t):
        print("Illegal character %s" % repr(t.value[0]))
        t.lexer.skip(1)

    #########
    # parser
    #########

    def p_prog(self, p):
        '''prog : DEF RUN LPAREN RPAREN LBRACE stmt RBRACE'''
        stmt = p[6]
        p[0] = lambda: stmt()

    def p_stmt(self, p):
        '''stmt : while
                | repeat
                | stmt_stmt
                | action
                | if
                | ifelse
        '''
        fn = p[1]
        p[0] = lambda: fn()

    def p_stmt_stmt(self, p):
        '''stmt_stmt : stmt SEMI stmt
        '''
        stmt1, stmt2 = p[1], p[3]
        def fn():
            stmt1(); stmt2();
        p[0] = fn

    def p_if(self, p):
        '''if : IF LPAREN cond RPAREN LBRACE stmt RBRACE
        '''
        cond, stmt = p[3], p[6]
        p[0] = lambda: stmt() if cond() else dummy()

    def p_ifelse(self, p):
        '''ifelse : IFELSE LPAREN cond RPAREN LBRACE stmt RBRACE ELSE LBRACE stmt RBRACE
        '''
        cond, stmt1, stmt2 = p[3], p[6], p[10]
        p[0] = lambda: stmt1() if cond() else stmt2()

    def p_while(self, p):
        '''while : WHILE LPAREN cond RPAREN LBRACE stmt RBRACE
        '''
        cond, stmt = p[3], p[6]
        def fn():
            while(cond()):
                stmt()
        p[0] = fn

    def p_repeat(self, p):
        '''repeat : REPEAT LPAREN cste RPAREN LBRACE stmt RBRACE
        '''
        cste, stmt = p[3], p[6]
        def fn():
            for _ in range(cste()):
                stmt()
        p[0] = fn

    def p_cond(self, p):
        '''cond : cond_without_not
                | NOT cond_without_not
        '''
        if callable(p[1]):
            cond_without_not = p[1]
            p[0] = lambda: cond_without_not()
        else: # NOT
            cond_without_not = p[2]
            p[0] = lambda: not cond_without_not()

    def p_cond_without_not(self, p):
        '''cond_without_not : FRONT_IS_CLEAR LPAREN RPAREN
                            | LEFT_IS_CLEAR LPAREN RPAREN
                            | RIGHT_IS_CLEAR LPAREN RPAREN
                            | MARKERS_PRESENT LPAREN RPAREN 
                            | NO_MARKERS_PRESENT LPAREN RPAREN
        '''
        cond_without_not = p[1]
        p[0] = lambda: getattr(self.karel, cond_without_not)()

    def p_action(self, p):
        '''action : MOVE LPAREN RPAREN
                  | TURN_RIGHT LPAREN RPAREN
                  | TURN_LEFT LPAREN RPAREN
                  | PICK_MARKER LPAREN RPAREN
                  | PUT_MARKER LPAREN RPAREN
        '''
        action = p[1]
        def fn():
            return getattr(self.karel, action)()
        p[0] = fn

    def p_cste(self, p):
        '''cste : INT
        '''
        value = p[1]
        p[0] = lambda: int(value)

    def p_error(self, p):
        if p:
            print("Syntax error at '%s'" % p.value)
        else:
            print("Syntax error at EOF")

    #########
    # Karel
    #########

    def get_state(self):
        return self.karel.state

    @timeout(0.001)
    def run(self, code, **kwargs):
        return yacc.parse(code, **kwargs)()

    def new_game(self, **kwargs):
        self.karel = Karel(debug=self.debug, **kwargs)

    def draw(self, *args, **kwargs):
        self.karel.draw(*args, **kwargs)

    def random_code(self, *args, **kwargs):
        return " ".join(self.random_tokens(*args, **kwargs))

    def random_tokens(self, start_token="prog", depth=0, stmt_max_depth=5):
        if start_token == 'stmt' and depth > stmt_max_depth:
            start_token = "action"

        codes = []
        candidates = self.prodnames[start_token]

        prod = candidates[self.rng.randint(len(candidates))]

        for term in prod.prod:
            if term in self.prodnames: # need digging
                codes.extend(self.random_tokens(term, depth + 1, stmt_max_depth))
            else:
                token = getattr(self, 't_{}'.format(term))
                if callable(token):
                    if token == self.t_INT:
                        token = self.random_INT()
                    else:
                        raise Exception(" [!] Undefined token `{}`".format(token))

                codes.append(str(token).replace('\\', ''))

        return codes

    def one_hot(self, string):
        pass

def test(parser, code=None):
    if code is None:
        code = parser.random_code()
        pprint(code)
    elif code.strip() == "":
        return

    parser.new_game(world_size=(8, 8))

    parser.draw("Input:  ")
    parser.run(code, debug=False)
    parser.draw("Output: ")



#!/usr/bin/env python


"""
Setup script for karel
"""
import os
try:
    from setuptools import setup
except ImportError:
    from distutils.core import setup

readme_content = "README: https://github.com/carpedm20/karel"

requirements = [
    'ply',
    'numpy',
]

version = None
author = None
email = None
source = None
description = None

with open(os.path.join('karel', '__init__.py')) as f:
    for line in f:
        if line.strip().startswith('__version__'):
            version = line.split('=')[1].strip().replace('"', '').replace("'", '')
        elif line.strip().startswith('__author__'):
            author = line.split('=')[1].strip().replace('"', '').replace("'", '')
        elif line.strip().startswith('__email__'):
            email = line.split('=')[1].strip().replace('"', '').replace("'", '')
        elif line.strip().startswith('__source__'):
            source = line.split('=')[1].strip().replace('"', '').replace("'", '')
        elif line.strip().startswith('__description__'):
            description = line.split('=')[1].strip().replace('"', '').replace("'", '')
        elif None not in (version, author, email, source, description):
            break

setup(
    name='karel',
    author=author,
    author_email=email,
    license='BSD License',
    keywords=["Interpreter and code generator of Karel"],
    description=description,
    long_description=readme_content,
    classifiers=[
        'Development Status :: 2 - Pre-Alpha',
        'Intended Audience :: Developers',
        'Intended Audience :: Developers',
        'Intended Audience :: Information Technology',
        'License :: OSI Approved :: BSD License',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
        'Programming Language :: Python :: 3.5',
        'Programming Language :: Python :: Implementation :: CPython',
        'Programming Language :: Python :: Implementation :: PyPy',
        'Programming Language :: Python',
        'Topic :: Software Development :: Libraries :: Python Modules',
        'Topic :: Internet :: WWW/HTTP :: Dynamic Content',
        'Topic :: Communications :: Chat',
    ],
    include_package_data=True,
    packages=['karel'],
    install_requires=requirements,
    url=source,
    version=version,
    zip_safe=True,
)



import os
import errno
import signal
import numpy as np
from functools import wraps
from pyparsing import nestedExpr

class TimeoutError(Exception):
    pass

def str2bool(v):
    return v.lower() in ('true', '1')

class Tcolors:
    CYAN = '\033[1;30m'
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'

def timeout(seconds=10, error_message=os.strerror(errno.ETIME)):
    def decorator(func):
        def _handle_timeout(signum, frame):
            raise TimeoutError(error_message)

        def wrapper(*args, **kwargs):
            signal.signal(signal.SIGALRM, _handle_timeout)
            signal.setitimer(signal.ITIMER_REAL,seconds) #used timer instead of alarm
            try:
                result = func(*args, **kwargs)
            finally:
                signal.alarm(0)
            return result
        return wraps(func)(wrapper)
    return decorator

def beautify_fn(inputs, indent=1, tabspace=2):
    lines, queue = [], []
    space = tabspace * " "

    for item in inputs:
        if item == ";":
            lines.append(" ".join(queue))
            queue = []
        elif type(item) == str:
            queue.append(item)
        else:
            lines.append(" ".join(queue + ["{"]))
            queue = []

            inner_lines = beautify_fn(item, indent=indent+1, tabspace=tabspace)
            lines.extend([space + line for line in inner_lines[:-1]])
            lines.append(inner_lines[-1])

    if len(queue) > 0:
        lines.append(" ".join(queue))

    return lines + ["}"]

def pprint(code, *args, **kwargs):
    print(beautify(code, *args, **kwargs))

def beautify(code, tabspace=2):
    array = nestedExpr('{','}').parseString("{"+code+"}").asList()
    lines = beautify_fn(array[0])
    return "\n".join(lines[:-1]).replace(' ( ', '(').replace(' )', ')')

def makedirs(path):
    if not os.path.exists(path):
        print(" [*] Make directories : {}".format(path))
        os.makedirs(path)

def get_rng(rng, seed=123):
    if rng is None:
        rng = np.random.RandomState(seed)
    return rng
